@@ -1,11 +1,11 @@
+use alloc::ffi::CString;
 use core::ptr::NonNull;
 
-use crate::__macro_helpers::declared_ivars::get_initialized_ivar_ptr;
+use crate::__macro_helpers::defined_ivars::get_initialized_ivar_ptr;
 use crate::encode::RefEncode;
-use crate::msg_send_id;
-use crate::mutability::{IsAllocableAnyThread, IsRetainable, Mutability};
 use crate::rc::{Allocated, Retained};
-use crate::runtime::{AnyClass, AnyProtocol};
+use crate::runtime::{AnyClass, AnyProtocol, ProtocolObject};
+use crate::MainThreadMarker;
 
 /// Types that can be sent Objective-C messages.
 ///
@@ -34,10 +34,10 @@ use crate::runtime::{AnyClass, AnyProtocol};
 /// runtime will do so. If you need to run some code when the object is
 /// destroyed, implement the `dealloc` method instead.
 ///
-/// The [`declare_class!`] macro does this for you, but the [`extern_class!`]
+/// The [`define_class!`] macro does this for you, but the [`extern_class!`]
 /// macro fundamentally cannot.
 ///
-/// [`declare_class!`]: crate::declare_class
+/// [`define_class!`]: crate::define_class
 /// [`extern_class!`]: crate::extern_class
 ///
 ///
@@ -80,7 +80,44 @@ use crate::runtime::{AnyClass, AnyProtocol};
 /// //
 /// // And `Retained<MyObject>` can now be constructed.
 /// ```
-pub unsafe trait Message: RefEncode {}
+///
+/// Implement the trait manually for a class with a lifetime parameter.
+///
+/// ```
+#[doc = include_str!("../examples/class_with_lifetime.rs")]
+/// ```
+pub unsafe trait Message: RefEncode {
+    /// Increment the reference count of the receiver.
+    ///
+    /// This extends the duration in which the receiver is alive by detaching
+    /// it from the lifetime information carried by the reference.
+    ///
+    /// This is similar to using [`Clone` on `Retained<Self>`][clone-id], with
+    /// the addition that it can be used on a plain reference.
+    ///
+    /// If your type may have come from a mutable type like `NSMutableString`,
+    /// you should consider using `NSCopying::copy` instead to avoid carrying
+    /// around a mutable string when you did not intend to.
+    ///
+    /// [clone-id]: crate::rc::Retained#impl-Clone-for-Retained<T>
+    #[inline]
+    #[doc(alias = "objc_retain")]
+    fn retain(&self) -> Retained<Self>
+    where
+        Self: Sized, // Temporary
+    {
+        let ptr: *const Self = self;
+        let ptr: *mut Self = ptr as _;
+        // SAFETY:
+        // - The pointer is valid since it came from `&self`.
+        // - The lifetime of the pointer itself is extended, but any lifetime
+        //   that the object may carry is still kept within the type itself.
+        let obj = unsafe { Retained::retain(ptr) };
+        // SAFETY: The pointer came from `&self`, which is always non-null,
+        // and objc_retain always returns the same value.
+        unsafe { obj.unwrap_unchecked() }
+    }
+}
 
 /// Marks types that represent specific classes.
 ///
@@ -90,30 +127,15 @@ pub unsafe trait Message: RefEncode {}
 /// Objective-C class - this trait allows you to communicate that, as well as
 /// a few properties of the class to the rest of the type-system.
 ///
-/// This is implemented automatically for your type by the
-/// [`declare_class!`][crate::declare_class] and
+/// This is implemented for your type by the
+/// [`define_class!`][crate::define_class] and
 /// [`extern_class!`][crate::extern_class] macros.
 ///
 ///
 /// # Safety
 ///
-/// 1. The type must represent a specific class.
-/// 2. [`Self::Super`] must be a superclass of the class (or something that
-///    represents any object, like [`AnyObject`][crate::runtime::AnyObject]).
-/// 3. [`Self::Mutability`] must be specified correctly.
-///
-///    Note that very little Objective-C code follows Rust's usual ownership
-///    model. If you think your type's mutability should be [`Mutable`], think
-///    again, it _very_ rarely should!
-///
-///    If you're unsure of what to do, [`InteriorMutable`] + avoiding `&mut`
-///    is usually a good starting point.
-/// 4. [`Self::NAME`] must be correct.
-/// 5. The class returned by [`Self::class`] must be the class that this type
-///    represents.
-///
-/// [`Mutable`]: crate::mutability::Mutable
-/// [`InteriorMutable`]: crate::mutability::InteriorMutable
+/// This is meant to be a sealed trait, and should not be implemented outside
+/// of the aforementioned macros. See those for safety preconditions.
 ///
 ///
 /// # Examples
@@ -121,7 +143,7 @@ pub unsafe trait Message: RefEncode {}
 /// Use the trait to access the [`AnyClass`] of an object.
 ///
 /// ```
-/// use objc2::{ClassType, msg_send_id};
+/// use objc2::{ClassType, msg_send};
 /// use objc2::rc::Retained;
 /// # use objc2::runtime::{NSObject as MyObject};
 ///
@@ -131,20 +153,20 @@ pub unsafe trait Message: RefEncode {}
 /// let cls = MyObject::class();
 ///
 /// // We can now access properties of the class.
-/// assert_eq!(cls.name(), MyObject::NAME);
+/// assert_eq!(cls.name().to_str().unwrap(), MyObject::NAME);
 ///
 /// // And we can send messages to the class.
 /// //
 /// // SAFETY:
 /// // - The class is `MyObject`, which can safely be initialized with `new`.
 /// // - The return type is correctly specified.
-/// let obj: Retained<MyObject> = unsafe { msg_send_id![cls, new] };
+/// let obj: Retained<MyObject> = unsafe { msg_send![cls, new] };
 /// ```
 ///
 /// Use the trait to allocate a new instance of an object.
 ///
 /// ```
-/// use objc2::{ClassType, msg_send_id};
+/// use objc2::{msg_send, AllocAnyThread};
 /// use objc2::rc::Retained;
 /// # use objc2::runtime::{NSObject as MyObject};
 ///
@@ -153,7 +175,7 @@ pub unsafe trait Message: RefEncode {}
 /// // Now we can call initializers on this newly allocated object.
 /// //
 /// // SAFETY: `MyObject` can safely be initialized with `init`.
-/// let obj: Retained<MyObject> = unsafe { msg_send_id![obj, init] };
+/// let obj: Retained<MyObject> = unsafe { msg_send![obj, init] };
 /// ```
 ///
 /// Use the [`extern_class!`][crate::extern_class] macro to implement this
@@ -161,29 +183,31 @@ pub unsafe trait Message: RefEncode {}
 ///
 /// ```
 /// use objc2::runtime::NSObject;
-/// use objc2::{extern_class, mutability, ClassType};
+/// use objc2::{extern_class, ClassType, AllocAnyThread};
 ///
 /// extern_class!(
+///     // SAFETY: The superclass is correctly specified, and the class can be
+///     // safely used from any thread.
+///     #[unsafe(super(NSObject))]
+///     # // For testing purposes
+///     # #[name = "NSObject"]
 ///     struct MyClass;
-///
-///     // SAFETY: The superclass and the mutability is correctly specified.
-///     unsafe impl ClassType for MyClass {
-///         type Super = NSObject;
-///         type Mutability = mutability::InteriorMutable;
-///         # // For testing purposes
-///         # const NAME: &'static str = "NSObject";
-///     }
 /// );
 ///
 /// let cls = MyClass::class();
 /// let obj = MyClass::alloc();
 /// ```
-///
-/// Implement the trait manually for a class with a lifetime parameter.
-///
-/// ```
-#[doc = include_str!("../examples/class_with_lifetime.rs")]
-/// ```
+//
+// Actual safety preconditions:
+//
+// 1. The type must represent a specific class.
+// 2. [`Self::Super`] must be a superclass of the class (or something that
+//    represents any object, like [`AnyObject`][crate::runtime::AnyObject]).
+// 3. [`Self::ThreadKind`] must be correct. It is safe to default to the
+//    super class' thread kind, `<Self::Super as ClassType>::ThreadKind`.
+// 4. [`Self::NAME`] must be the name of the class that this type represents.
+// 5. The class returned by [`Self::class`] must be the class that this type
+//    represents.
 pub unsafe trait ClassType: Message {
     /// The superclass of this class.
     ///
@@ -197,15 +221,25 @@ pub unsafe trait ClassType: Message {
     /// [`AnyObject`]: crate::runtime::AnyObject
     type Super: Message;
 
-    /// Whether the type is mutable or immutable.
+    /// Whether the type can be used from any thread, or from only the main
+    /// thread.
     ///
-    /// See the [`mutability`][crate::mutability] module for further details
-    /// about class mutability.
-    type Mutability: Mutability;
+    /// One of [`dyn AllocAnyThread`] or [`dyn MainThreadOnly`].
+    ///
+    /// Setting this makes `ClassType` provide an implementation of either
+    /// [`AllocAnyThread`] or [`MainThreadOnly`].
+    ///
+    /// [`dyn AllocAnyThread`]: AllocAnyThread
+    /// [`dyn MainThreadOnly`]: MainThreadOnly
+    type ThreadKind: ?Sized + ThreadKind;
 
     /// The name of the Objective-C class that this type represents.
     ///
     /// `T::NAME` is the `const` version of `T::class().name()`.
+    ///
+    /// This must not contain any NUL bytes.
+    //
+    // TODO: Convert this to CStr next time we do big changes to ClassType.
     const NAME: &'static str;
 
     /// Get a reference to the Objective-C class that this type represents.
@@ -215,7 +249,7 @@ pub unsafe trait ClassType: Message {
     ///
     /// # Panics
     ///
-    /// This may panic if something went wrong with getting or declaring the
+    /// This may panic if something went wrong with getting or creating the
     /// class, e.g. if the program is not properly linked to the framework
     /// that defines the class.
     fn class() -> &'static AnyClass;
@@ -225,103 +259,29 @@ pub unsafe trait ClassType: Message {
     // we wanted to!
     fn as_super(&self) -> &Self::Super;
 
-    /// Get a mutable reference to the superclass.
-    // Note: No `Self: IsMutable` bound required here, since there is no way
-    // to get `&mut self` in the first place.
-    //
-    // Or at least, if we have `&mut NSMutableString`, we're allowed to get
-    // `&mut NSString`, and from that it will also make sense to allow getting
-    // `&mut NSObject`.
-    fn as_super_mut(&mut self) -> &mut Self::Super;
+    #[doc(hidden)]
+    const __INNER: ();
 
-    /// Increment the reference count of the receiver.
+    /// Inner type to use when subclassing with `define_class!`.
     ///
-    /// This extends the duration in which the receiver is alive by detaching
-    /// it from the lifetime information carried by the reference.
-    ///
-    /// This is similar to using [`Clone` on `Retained<Self>`][clone-id], with
-    /// the addition that it can be used on a plain reference. Note however
-    /// that this is not possible to use on certain types like `NSString`,
-    /// since if you only hold `&NSString`, that may have come from
-    /// `&mut NSMutableString`, in which case it would be unsound to erase the
-    /// lifetime information carried by the reference.
-    ///
-    /// In cases like that, you should rather use `NSCopying::copy` (since
-    /// that gives you a `NSString` whether the string was originally a
-    /// `NSString` or a `NSMutableString`).
-    ///
-    /// [clone-id]: crate::rc::Retained#impl-Clone-for-Retained<T>
-    //
-    // Note: We could have placed this on `mutability::IsRetainable`, but
-    // `ClassType` is more often already in scope, allowing easier access to
-    // `obj.retain()`.
-    #[inline]
-    #[doc(alias = "objc_retain")]
-    fn retain(&self) -> Retained<Self>
-    where
-        Self: IsRetainable,
-        Self: Sized, // Temporary
-    {
-        let ptr: *const Self = self;
-        let ptr: *mut Self = ptr as _;
-        // SAFETY:
-        // - The object is known to not be mutable (or have a mutable
-        //   subclass) due to the `IsRetainable` bound.
-        // - The pointer is valid since it came from `&self`.
-        // - The lifetime of the pointer itself is extended, but any lifetime
-        //   that the object may carry is still kept within the type itself.
-        let obj = unsafe { Retained::retain(ptr) };
-        // SAFETY: The pointer came from `&self`, which is always non-null
-        // (and objc_retain always returns the same value).
-        unsafe { obj.unwrap_unchecked() }
-    }
-
-    /// Allocate a new instance of the class.
-    ///
-    /// The return value can be used directly inside [`msg_send_id!`] to
-    /// initialize the object.
-    ///
-    /// For classes that are only usable on the main thread, you can use
-    /// `MainThreadMarker::alloc` instead.
-    ///
-    /// [`msg_send_id!`]: crate::msg_send_id
-    //
-    // Note: We could have placed this on `mutability::IsAllocableAnyThread`,
-    // but `ClassType` is more often already in scope, allowing easier access
-    // to `T::alloc()`.
-    #[inline]
-    fn alloc() -> Allocated<Self>
-    where
-        Self: IsAllocableAnyThread + Sized,
-    {
-        // SAFETY:
-        // - It is always safe to (attempt to) allocate an object.
-        // - The object is of the correct type, since we've used the class
-        //   from `Self::class`.
-        // - The object is safe to `dealloc` on the current thread (due to the
-        //   `IsAllocableAnyThread` bound which guarantees it is not
-        //   `MainThreadOnly`).
-        //
-        // See also `MainThreadMarker::alloc`.
-        unsafe { msg_send_id![Self::class(), alloc] }
-    }
-
-    // TODO: `fn alloc_on_main(mtm: MainThreadMarker)`
-    // TODO: `fn mtm(&self) -> MainThreadMarker where T::Mutability: MainThreadOnly`
+    /// This is used by NSObject to control which auto traits are set for
+    /// defined subclasses. Set to `= Self` in all other cases.
+    #[doc(hidden)]
+    type __SubclassingType: ?Sized;
 }
 
-/// Marks types whose implementation is defined in Rust.
+/// Marks class types whose implementation is defined in Rust.
 ///
-/// This is used in [`declare_class!`], and allows access to the instance
+/// This is used in [`define_class!`], and allows access to the instance
 /// variables that a given type declares, see that macro for details.
 ///
-/// [`declare_class!`]: crate::declare_class
+/// [`define_class!`]: crate::define_class
 //
 // Note: We mark this trait as not `unsafe` for better documentation, since
-// implementing it inside `declare_class!` is not `unsafe`.
+// implementing it inside `define_class!` is not `unsafe`.
 //
 // Safety is ensured by `__UNSAFE_OFFSETS_CORRECT`.
-pub trait DeclaredClass: ClassType {
+pub trait DefinedClass: ClassType {
     /// A type representing the instance variables that this class carries.
     type Ivars: Sized;
 
@@ -342,25 +302,6 @@ pub trait DeclaredClass: ClassType {
         unsafe { ivars.as_ref() }
     }
 
-    /// Get a mutable reference to the instance variable data that this object
-    /// carries.
-    #[inline]
-    #[track_caller]
-    fn ivars_mut(&mut self) -> &mut Self::Ivars
-    where
-        Self: Sized, // Required because of MSRV
-    {
-        let ptr: NonNull<Self> = NonNull::from(self);
-        // SAFETY: The pointer is valid and initialized.
-        let mut ivars = unsafe { get_initialized_ivar_ptr(ptr) };
-        // SAFETY: The lifetime of the instance variable is tied to the object.
-        //
-        // Mutability is safe since the object itself is mutable. See
-        // `ClassType::as_super_mut` for why this is safe without
-        // `Self: IsMutable`.
-        unsafe { ivars.as_mut() }
-    }
-
     #[doc(hidden)]
     fn __ivars_offset() -> isize;
 
@@ -400,25 +341,28 @@ pub trait DeclaredClass: ClassType {
 /// use objc2::runtime::NSObjectProtocol;
 /// // Get a protocol object representing the `NSObject` protocol
 /// let protocol = <dyn NSObjectProtocol>::protocol().expect("NSObject to have a protocol");
-/// assert_eq!(<dyn NSObjectProtocol>::NAME, protocol.name());
+/// assert_eq!(<dyn NSObjectProtocol>::NAME, protocol.name().to_str().unwrap());
 /// # // Ensure Foundation links on GNUStep
 /// # let _cls = objc2::class!(NSObject);
 /// ```
 ///
-/// Use the [`extern_protocol!`] macro to implement this trait for a type.
+/// Use the [`extern_protocol!`] macro to implement and use this trait.
 ///
 /// ```no_run
 /// use objc2::{extern_protocol, ProtocolType};
 ///
 /// extern_protocol!(
 ///     unsafe trait MyProtocol {}
-///     unsafe impl ProtocolType for dyn MyProtocol {}
 /// );
 ///
 /// let protocol = <dyn MyProtocol>::protocol();
 /// ```
 pub unsafe trait ProtocolType {
     /// The name of the Objective-C protocol that this type represents.
+    ///
+    /// This must not contain any NUL bytes.
+    //
+    // TODO: Convert this to CStr next time we do big changes to ProtocolType.
     const NAME: &'static str;
 
     /// Get a reference to the Objective-C protocol object that this type
@@ -435,13 +379,220 @@ pub unsafe trait ProtocolType {
     ///
     /// # Panics
     ///
-    /// This may panic if something went wrong with getting or declaring the
+    /// This may panic if something went wrong with getting or creating the
     /// protocol, e.g. if the program is not properly linked to the framework
     /// that defines the protocol.
     fn protocol() -> Option<&'static AnyProtocol> {
-        AnyProtocol::get(Self::NAME)
+        get_protocol(Self::NAME)
     }
 
     #[doc(hidden)]
     const __INNER: ();
 }
+
+// Outlined to reduce code size
+fn get_protocol(name: &str) -> Option<&'static AnyProtocol> {
+    let name = CString::new(name).expect("protocol name must be UTF-8");
+    AnyProtocol::get(&name)
+}
+
+// Split into separate traits for better diagnostics
+mod private {
+    pub trait SealedAllocAnyThread {}
+    pub trait SealedMainThreadOnly {}
+    pub trait SealedThreadKind {}
+}
+
+/// Marker trait for classes (and protocols) that are usable from any thread,
+/// i.e. the opposite of [`MainThreadOnly`].
+///
+/// This is mostly an implementation detail to expose the [`alloc`] method
+/// with different signatures depending on whether a class is main thread only
+/// or not. You can safely assume that things are safe to use from any thread,
+/// _unless_ they implement [`MainThreadOnly`], not only if they implement
+/// this trait.
+///
+///
+/// # Safety
+///
+/// This is a sealed trait, and should not need to be implemented; it is
+/// implemented automatically when you implement [`ClassType`].
+//
+// NOTE: Ideally this would be an auto trait that had a negative impl for
+// `MainThreadOnly`, something like:
+//
+//     pub unsafe auto trait AnyThread {}
+//     pub unsafe trait MainThreadOnly {}
+//     impl<T: ?Sized + MainThreadOnly> !AnyThread for T {}
+//
+// This isn't possible in current Rust though, so we'll have to hack it.
+pub unsafe trait AllocAnyThread: private::SealedAllocAnyThread {
+    /// Allocate a new instance of the class.
+    ///
+    /// The return value can be used directly inside [`msg_send!`] to
+    /// initialize the object.
+    ///
+    /// [`msg_send!`]: crate::msg_send
+    #[inline]
+    fn alloc() -> Allocated<Self>
+    where
+        Self: Sized + ClassType,
+    {
+        // SAFETY:
+        // - It is always safe to (attempt to) allocate an object.
+        // - The object is of the correct type, since we've used the class
+        //   from `Self::class`.
+        // - The object is safe to `dealloc` on the current thread (due to the
+        //   `AnyThread` bound which guarantees it is not `MainThreadOnly`).
+        //
+        // While Xcode's Main Thread Checker doesn't report `alloc` and
+        // `dealloc` as unsafe from other threads, things like `NSView` and
+        // `NSWindow` still do a non-trivial amount of stuff on `dealloc`,
+        // even if the object is freshly `alloc`'d - which is why we disallow
+        // this.
+        //
+        // This also has the nice property that `Allocated<T>` is guaranteed
+        // to be allowed to `init` on the current thread.
+        //
+        // See also `MainThreadMarker::alloc`.
+        unsafe { Allocated::alloc(Self::class()) }
+    }
+}
+
+// The impl here is a bit bad for diagnostics, but required to prevent users
+// implementing the trait themselves.
+impl<'a, T: ?Sized + ClassType<ThreadKind = dyn AllocAnyThread + 'a>> private::SealedAllocAnyThread
+    for T
+{
+}
+unsafe impl<'a, T: ?Sized + ClassType<ThreadKind = dyn AllocAnyThread + 'a>> AllocAnyThread for T {}
+
+impl<P: ?Sized> private::SealedAllocAnyThread for ProtocolObject<P> {}
+unsafe impl<P: ?Sized + AllocAnyThread> AllocAnyThread for ProtocolObject<P> {}
+
+/// Marker trait for classes and protocols that are only safe to use on the
+/// main thread.
+///
+/// This is commonly used in GUI code like `AppKit` and `UIKit`, e.g.
+/// `UIWindow` is only usable from the application's main thread because it
+/// accesses global statics like the `UIApplication`.
+///
+/// See [`MainThreadMarker`] for a few more details on this.
+///
+///
+/// # Safety
+///
+/// It is unsound to implement [`Send`] or [`Sync`] together with this.
+///
+/// This is a sealed trait, and should not need to be implemented; it is
+/// implemented automatically when you implement [`ClassType`].
+#[doc(alias = "@MainActor")]
+pub unsafe trait MainThreadOnly: private::SealedMainThreadOnly {
+    /// Get a [`MainThreadMarker`] from the main-thread-only object.
+    ///
+    /// This function exists purely in the type-system, and will succeed at
+    /// runtime (with a safety check when debug assertions are enabled).
+    #[inline]
+    #[cfg_attr(debug_assertions, track_caller)]
+    fn mtm(&self) -> MainThreadMarker {
+        #[cfg(debug_assertions)]
+        assert!(
+            MainThreadMarker::new().is_some(),
+            "the main-thread-only object that we tried to fetch a MainThreadMarker from was somehow not on the main thread",
+        );
+
+        // SAFETY: Objects which are `MainThreadOnly` are guaranteed
+        // `!Send + !Sync` and are only constructible on the main thread.
+        //
+        // Since we hold `&self`, i.e. a reference to such an object, and we
+        // know it cannot possibly be on another thread than the main, we know
+        // that the current thread is the main thread.
+        unsafe { MainThreadMarker::new_unchecked() }
+    }
+
+    /// Allocate a new instance of the class on the main thread.
+    ///
+    ///
+    /// # Example
+    ///
+    /// Create a view on the main thread.
+    ///
+    /// ```
+    /// use objc2::{MainThreadOnly, MainThreadMarker};
+    /// # #[cfg(available_in_app_kit)]
+    /// use objc2_app_kit::NSView;
+    /// use objc2_core_foundation::CGRect;
+    /// #
+    /// # use objc2::rc::{Allocated, Retained};
+    /// #
+    /// # objc2::extern_class!(
+    /// #     #[unsafe(super(objc2::runtime::NSObject))]
+    /// #     #[thread_kind = MainThreadOnly]
+    /// #     #[name = "NSObject"] // For example
+    /// #     struct NSView;
+    /// # );
+    /// #
+    /// # impl NSView {
+    /// #     fn initWithFrame(this: Allocated<Self>, _frame: CGRect) -> Retained<Self> {
+    /// #         // Don't use frame, this is NSObject
+    /// #         unsafe { objc2::msg_send![this, init] }
+    /// #     }
+    /// # }
+    ///
+    /// # #[cfg(doctests_not_always_run_on_main_thread)]
+    /// let mtm = MainThreadMarker::new().expect("must be on the main thread");
+    /// # let mtm = unsafe { MainThreadMarker::new_unchecked() };
+    ///
+    /// let frame = CGRect::default();
+    /// let view = NSView::initWithFrame(NSView::alloc(mtm), frame);
+    /// ```
+    #[inline]
+    fn alloc(mtm: MainThreadMarker) -> Allocated<Self>
+    where
+        Self: Sized + ClassType,
+    {
+        let _ = mtm;
+        // SAFETY: We hold `MainThreadMarker`, and the class is safe to
+        // allocate on the main thread.
+        unsafe { Allocated::alloc(Self::class()) }
+    }
+}
+
+impl<'a, T: ?Sized + ClassType<ThreadKind = dyn MainThreadOnly + 'a>> private::SealedMainThreadOnly
+    for T
+{
+}
+unsafe impl<'a, T: ?Sized + ClassType<ThreadKind = dyn MainThreadOnly + 'a>> MainThreadOnly for T {}
+
+impl<P: ?Sized> private::SealedMainThreadOnly for ProtocolObject<P> {}
+unsafe impl<P: ?Sized + MainThreadOnly> MainThreadOnly for ProtocolObject<P> {}
+
+/// The allowed values in [`ClassType::ThreadKind`].
+///
+/// One of [`dyn AllocAnyThread`] or [`dyn MainThreadOnly`].
+///
+/// [`dyn AllocAnyThread`]: AllocAnyThread
+/// [`dyn MainThreadOnly`]: MainThreadOnly
+pub trait ThreadKind: private::SealedThreadKind {
+    // To mark `ThreadKind` as dyn-incompatible for now.
+    #[doc(hidden)]
+    const __DYN_INCOMPATIBLE: ();
+}
+
+impl private::SealedThreadKind for dyn AllocAnyThread + '_ {}
+impl ThreadKind for dyn AllocAnyThread + '_ {
+    const __DYN_INCOMPATIBLE: () = ();
+}
+
+impl private::SealedThreadKind for dyn MainThreadOnly + '_ {}
+impl ThreadKind for dyn MainThreadOnly + '_ {
+    const __DYN_INCOMPATIBLE: () = ();
+}
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    #[allow(unused)]
+    fn dyn_compatible(_: &dyn AllocAnyThread, _: &dyn MainThreadOnly) {}
+}
